/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <linux/soundcard.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <errno.h>
#include "../gdamserverdebug.h"
#include "gdamdeviceoss.h"

#define GDAM_SOURCE_PRIORITY G_PRIORITY_HIGH

typedef struct _GdamDeviceOptions GdamDeviceOptions;

static gboolean    gdam_device_oss_open      (GdamSink*         sink,
                                              const char*       name,
				              const char*       opts);
static void        gdam_device_oss_close     (GdamSink*         sink);
static gboolean    gdam_device_oss_set_source(GdamSink*         sink,
                                              GdamSource*       source);
static void        gdam_device_oss_remove_source(GdamSink*      sink);
static int         gdam_device_oss_write     (GdamSink*         output,
	                                      const GdamSample* samples,
					      int               max_read);
static void        gdam_device_oss_destroy   (GtkObject*        to_kill);


static int         gdam_device_oss_open_fd  (const char*        device_name,
                                             GdamDeviceOptions* device_opts);

/* Add/remove the sink from the main loop. */
static void        gdam_device_oss_stop      (GdamDeviceOss*    device);
static void        gdam_device_oss_play      (GdamDeviceOss*    device);

/* Functions passed to g_source_add. */
static gboolean    gdam_device_oss_prepare   (gpointer          src_data,
                                              GTimeVal*         curtime,
                                              gint*             timeout,
		                              gpointer          user_data);
static gboolean    gdam_device_oss_check     (gpointer          src_data,
                                              GTimeVal*         curtime,
		                              gpointer          user_data);
static gboolean    gdam_device_oss_dispatch  (gpointer          src_data,
                                              GTimeVal*         curtime,
		                              gpointer          user_data);

/* And this function does a single fragment-sized write out to
 * the file descriptor. */
static gboolean    gdam_device_oss_write_one_fragment
                                             (GdamDeviceOss*    device_oss);


static void gdam_device_oss_class_init(GdamSinkClass*  klass)
{
	GtkObjectClass* object_class = (GtkObjectClass*)klass;
	klass->write = gdam_device_oss_write;
	klass->set_source = gdam_device_oss_set_source;
	klass->remove_source = gdam_device_oss_remove_source;
	klass->open = gdam_device_oss_open;
	klass->close = gdam_device_oss_close;
	object_class->destroy = gdam_device_oss_destroy;
}

static void gdam_device_oss_init(GdamDeviceOss*  sink)
{
	sink->device_fd = -1;
	sink->poll_fd.fd = -1;
	sink->source_id = 0;
	sink->fragment_size = (1 << 11);
}

static GtkObjectClass* the_sink_class;
static GtkType device_oss_type = 0;
GtkType gdam_device_oss_get_type()
{
	if (!device_oss_type) {
		static const GtkTypeInfo sink_info = {
			"GdamDeviceOss",
			sizeof(GdamDeviceOss),
			sizeof(GdamSinkClass),
			(GtkClassInitFunc) gdam_device_oss_class_init,
			(GtkObjectInitFunc) gdam_device_oss_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		the_sink_class = gtk_type_class(GDAM_TYPE_SINK);
		device_oss_type = gtk_type_unique(GDAM_TYPE_SINK, &sink_info);
	}
	return device_oss_type;
}

GdamSink* gdam_device_oss_new(gpointer      device_data,
                              const char*   device_name,
			      const char*   device_options)
{
	GtkType type = gdam_device_oss_get_type();
	GdamDeviceOss* dev_oss = (GdamDeviceOss*)gtk_type_new(type);
	if (!gdam_sink_open(GDAM_SINK(dev_oss), device_name, device_options)) {
		gtk_object_unref(GTK_OBJECT(dev_oss));
		return NULL;
	}
	return GDAM_SINK(dev_oss);
}

#define GDAM_SAMPLE_FORMAT_NATIVE16	0
#define GDAM_SAMPLE_FORMAT_NATIVE8	1
struct _GdamDeviceOptions {
	int		rate;		/* samples per second. */
	gboolean	do_ioctls;	/* Whether or not to actually 
	                                 *    run the sound card ioctls.
					 * Needed to reuse this class for
					 *    raw output. */
	int		num_channels;	/* must be 2 for now. */
	int		format;

	int		num_fragments;
	int		log_frag_size;

	gboolean	should_fork;

	void           (*downsample_func)(void* samples, int num);
	int              sample_size;		/* in bits. */
};

/*=====================================================================
 * Downsampling info...
 *
 * TODO:  move this into its own file. */


typedef struct _DownsampleInfo DownsampleInfo;
struct _DownsampleInfo {
	int rate;
	int stereo;
	int output_size;
	int sample_size;		/* in bits, compensating rate */
	void (*downsample_func)(GdamSample* in, int);
};

static void downsample_16mono44100(GdamSample* buf, int n)
{
	gint16* out = (gint16*)buf;
	while (n-- > 0) {
		*out = ((int)buf->s[0] + buf->s[1]) / 2;
		buf++; out++;
	}
}
static void downsample_8mono44100(GdamSample* buf, int n)
{
	guint8* out = (guint8*)buf;
	while (n-- > 0) {
		int avg = ((int)buf->s[0] + buf->s[1]) / 2;
		*out = (guint8)(avg>>8) ^ 0x80;
		buf++; out++;
	}
}
static void downsample_8stereo44100(GdamSample* buf, int n)
{
	guint8* out = (guint8*)buf;
	while (n-- > 0) {
		*out++ = (guint8)(buf->s[0]>>8) ^ 0x80;
		*out++ = (guint8)(buf->s[1]>>8) ^ 0x80;
		buf++;
	}
}
static void downsample_16mono22050(GdamSample* buf, int n)
{
	gint16* out = (gint16*)buf;
	n /= 2;
	while (n-- > 0) {
		*out = (gint16) ((int)buf[0].s[0] + buf[0].s[1]
		                    + buf[1].s[0] + buf[1].s[1]) / 4;
		buf += 2; out++;
	}
}
static void downsample_8mono22050(GdamSample* buf, int n)
{
	guint8* out = (guint8*)buf;
	n /= 2;
	while (n-- > 0) {
		*out = (guint8) (((int)buf[0].s[0] + buf[0].s[1]
		                     + buf[1].s[0] + buf[1].s[1])/4/256);
		*out ^= 0x80;
		buf += 2; out++;
	}
}
static void downsample_16stereo22050(GdamSample* buf, int n)
{
	GdamSample* out = buf;
	n /= 2;
	while (n-- > 0) {
		out->s[0] = (gint16) (((int)buf[0].s[0] + buf[1].s[0]) / 2);
		out->s[1] = (gint16) (((int)buf[0].s[1] + buf[1].s[1]) / 2);
		buf += 2; out++;
	}
}
static void downsample_8stereo22050(GdamSample* buf, int n)
{
	guint8* out = (guint8*)buf;
	n /= 2;
	while (n-- > 0) {
		*out++ = (guint8) (((int)buf[0].s[0] + buf[1].s[0])/2/256)
						^ 0x80;
		*out++ = (guint8) (((int)buf[0].s[1] + buf[1].s[1])/2/256)
						^ 0x80;
		buf += 2;
	}
}

static DownsampleInfo methods[] = {
	{  44100,       1,   16,   32,                      NULL },
	{  44100,       1,    8,   16,   downsample_8stereo44100 },
	{  44100,   FALSE,   16,   16,    downsample_16mono44100 },
	{  44100,   FALSE,    8,    8,     downsample_8mono44100 },

	{  22050,       1,   16,   16,  downsample_16stereo22050 },
	{  22050,       1,    8,    8,   downsample_8stereo22050 },
	{  22050,   FALSE,   16,    8,    downsample_16mono22050 },
	{  22050,   FALSE,    8,    4,     downsample_8mono22050 },
};

#define NUM_METHODS ((sizeof(methods) / sizeof(methods[0])))

static gboolean get_downsampler(GdamDeviceOptions* options, 
                                int                rate,
				int                stereo,
				int                sample_size)
{
	int i;
	stereo = stereo ? 1 : 0;
	for (i = 0; i < NUM_METHODS; i++) {
		if ( (rate - 1000 < methods[i].rate
		   && methods[i].rate < rate + 1000)
		 && methods[i].stereo == stereo
		 && methods[i].output_size == sample_size) {
		 	options->downsample_func
				= (void (*)(void*,int))
						methods[i].downsample_func;
			options->sample_size = methods[i].sample_size;
			return TRUE;
		}
	}
	g_warning("no downsample method for %d hz, %s, %d bits/sample",
		rate, stereo ? "stereo" : "mono", sample_size);
	return FALSE;
}


/*
 *                                  _            _
 *  _ __   __ _ _ __ ___  ___    __| | _____   _(_) ___ ___
 * | '_ \ / _` | '__/ __|/ _ \  / _` |/ _ \ \ / / |/ __/ _ \
 * | |_) | (_| | |  \__ \  __/ | (_| |  __/\ V /| | (_|  __/
 * | .__/ \__,_|_|  |___/\___|  \__,_|\___| \_/ |_|\___\___|
 * |_|
 *              _   _
 *   ___  _ __ | |_(_) ___  _ __  ___
 *  / _ \| '_ \| __| |/ _ \| '_ \/ __|
 * | (_) | |_) | |_| | (_) | | | \__ \
 *  \___/| .__/ \__|_|\___/|_| |_|___/
 *       |_|
 */
static gboolean gdam_device_parse_options(GdamDeviceOptions* options,
                                          const char*        string)
{
	char** opts;
	char** opts_at;
	options->rate = 44100;
	options->do_ioctls = TRUE;
	options->num_channels = 2;
	options->format = GDAM_SAMPLE_FORMAT_NATIVE16;
	options->num_fragments = 4;
	options->log_frag_size = 11;
	options->should_fork = FALSE;
	opts = g_strsplit(string, ":", 100);
	for (opts_at = opts; *opts_at; opts_at++) {
		char* opt = *opts_at;
		if (strncmp(opt, "logfragsize=", 12) == 0) {
			int log_frag_size = atoi(opt + 12);
			if (log_frag_size < 6 || log_frag_size > 100) {
				g_warning("log_frag_size was out of bounds...");
			} else {
				options->log_frag_size = log_frag_size;
			}
		} else if (strncmp(opt, "bits=", 5) == 0) {
			int bits = atoi(opt + 5);
			if (bits == 8)
				options->format = GDAM_SAMPLE_FORMAT_NATIVE8;
			else if (bits == 16)
				options->format = GDAM_SAMPLE_FORMAT_NATIVE16;
			else {
				g_warning("unknown bits/sample %d", bits);
			}
		} else if (strcmp(opt, "mono") == 0) {
			options->num_channels = 1;
		} else if (strcmp(opt, "stereo") == 0) {
			options->num_channels = 2;
		} else if (strncmp(opt, "rate=", 5) == 0) {
			options->rate = atoi(opt + 5);
		} else if (strcmp(opt, "fork") == 0) {
			options->should_fork = TRUE;
		} else if (strncmp(opt, "numfragments=", 13) == 0) {
			int num_fragments = atoi(opt + 13);
			if (num_fragments < 3 || num_fragments > 300) {
				g_warning("numfragments was out of bounds...");
			} else {
				options->num_fragments = num_fragments;
			}
		} else {
			g_warning("only options known to oss are "
				"numfragments and logfragsize [was %s]", opt);
		}
		g_free(opt);
	}
        options->sample_size = options->format == GDAM_SAMPLE_FORMAT_NATIVE16
	                                  ? 16 : 8;
        options->sample_size *= options->num_channels;

	g_free(opts);

	return TRUE;
}
static int bits_per_sample(int format)
{
	if (format == GDAM_SAMPLE_FORMAT_NATIVE16)	return 16;
	if (format == GDAM_SAMPLE_FORMAT_NATIVE8)	return 8;
	return -1;
}

static gboolean writen(int fd, const char* buf, int len)
{
        while (len > 0) {
                int n = write(fd, buf, len);
                if (n < 0) {
                        if (errno == EINTR)
                                continue;
                        return FALSE;
                }
                buf += n; len -= n;
        }
        return TRUE;
}

static gboolean readn(int fd, char* buf, int len)
{
        while (len > 0) {
                int n = read(fd, buf, len);
                if (n < 0) {
                        if (errno == EINTR)
                                continue;
                        return FALSE;
                }
                if (n == 0) return FALSE;
                buf += n; len -= n;
        }
        return TRUE;
}


static int fork_open_fd(const char*        device_name,
                         GdamDeviceOptions* device_options)
{
	int pid;
	int fds[2];

	int soundcard_fd;
	char buffer[4096];

	if (pipe(fds) < 0) {
		perror("opening pipe");
		exit(1);
	}
	while ((pid = fork()) < 0) {
		if (errno == EINTR) continue;
		if (errno == EAGAIN) { sleep(3); continue; }
		perror("forking");
		exit(1);
	}


	if (pid != 0) {
		close(fds[0]);
		return fds[1];
	}
	close(fds[1]);
	/* Child process. */
	soundcard_fd = gdam_device_oss_open_fd(device_name, device_options);
	while (readn(fds[0], buffer, sizeof(buffer))
	  &&   writen(soundcard_fd, buffer, sizeof(buffer)))
	  	;
	_exit(0);
}

static int
gdam_device_oss_open_fd(const char*          device_name,
                        GdamDeviceOptions*   device_options)
{
	int fd;
	int t; unsigned long r;
	int log_frag_size = device_options->log_frag_size;
	int num_fragments = device_options->num_fragments;
	int stereo = 1, rate, sample_size, num_channels;
	if (!device_name)
		device_name = "/dev/dsp";
	if (device_options->should_fork) {
		device_options->should_fork = FALSE;
		gdam_debug("forking and opening %s", device_name);
		fd = fork_open_fd(device_name, device_options);
		device_options->should_fork = TRUE;
		device_options->num_channels = 2;
		device_options->format = GDAM_SAMPLE_FORMAT_NATIVE16;
		device_options->sample_size = 16 * device_options->num_channels;
		device_options->downsample_func = NULL;
		return fd;
	}

	fd = open(device_name, O_WRONLY, 0);
	if (fd < 0) {
		fprintf(stderr, "couldn't open sound device %s: %s\n",
			device_name, strerror(errno));
		return -1;
	}
	if (!device_options->do_ioctls)
		return fd;

	/* Close the device on exec. */
	fcntl(fd, F_SETFD, FD_CLOEXEC);

	t = (log_frag_size & 0xffff) | ((num_fragments & 0xffff)<<16);
	if (ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &t) < 0) {
		g_warning("error going to correct amount of frag (%s)",
				strerror(errno));
		goto err;
	} else {
		int total_bytes = num_fragments << log_frag_size;
		gdam_debug("Set soundcard fragments:  "
			   "%d frags of %d bytes; %3.3fs",
			   	num_fragments,
				(1<<log_frag_size),
		                (float)total_bytes/(44100*sizeof(GdamSample)));
	}
	t = bits_per_sample(device_options->format);
	if (t < 0) {
		g_warning("unknown sample format %d", device_options->format);
		goto err;
	}
	if (ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &t) < 0) {
		g_warning("error going to %d bits per sample", t);
		goto err;
	}
	if (device_options->num_channels == 1) {
		t = 0;		/* non-stereo */
	} else if (device_options->num_channels == 2) {
		t = 1;		/* stereo */
	} else {
		g_warning("only numbers of channels allowed are "
			  "1 (mono) or 2 (stereo), num_channels was %d",
			  device_options->num_channels);
		goto err;
	}
		
	if (ioctl(fd, SNDCTL_DSP_STEREO, &t) < 0) {
		g_warning("error setting device to stereo: %s",
			  strerror(errno));
		goto err;
	}
#ifdef SNDCTL_DSP_PROFILE
	t = APF_NETWORK;
	if (ioctl(fd, SNDCTL_DSP_PROFILE, &t) < 0) {
		g_warning("unable to tell soundcard bufferer "
			"to expect pops from us");
	}
#else
	g_warning("not compiled with DSP_PROFILE support");
#endif
	r = device_options->rate;
#if 0
	if (r != 44100) {
		g_warning("only supported output rate is %d, "
			  "you are using %lu,"
		          " but go ahead...", 44100, r);
	}
#endif
	if (ioctl(fd, SNDCTL_DSP_SPEED, &r) < 0) {
		g_warning("error setting dev to %d samples per sec", t);
		goto err;
	}
	if (ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &r) < 0) {
		g_warning("error getting dev block size");
		goto err;
	}
	gdam_debug("DSP Block size is %lu", r);
	/*sound_device_block_size = r;*/

	rate = num_channels = sample_size = -1;
	if (ioctl(fd, SOUND_PCM_READ_RATE, &rate) < 0
	 || ioctl(fd, SOUND_PCM_READ_BITS, &sample_size) < 0
	 || ioctl(fd, SOUND_PCM_READ_CHANNELS, &num_channels) < 0
	 || rate == -1
	 || num_channels == -1
	 || sample_size == -1) {
                g_warning("error getting device parameters");
		g_warning("you may need to manually specify the soundcard "
			  "rate if it does not support the defaults");
                rate = device_options->rate;
                num_channels = device_options->num_channels;
                if (num_channels <= 0) num_channels = 1;
                sample_size = device_options->sample_size / num_channels;
	}
	if (num_channels == 1)          stereo = 0;
	else if (num_channels == 2)     stereo = 1;
	else {
		g_error("unsupported number of channels %d", num_channels);
	}
	gdam_debug("Device set at %d bits/sample, stereo %s, rate %d",
			sample_size, stereo ? "on" : "off", rate);
	
	if (!get_downsampler(device_options,
	                     rate,
			     stereo,
			     sample_size)) {
		g_error("no way to convert samples for your device");
	}

	return fd;
err:
	close(fd);
	return -1;
}

/* Open the sound device. */
static gboolean
gdam_device_oss_open(GdamSink* sink, const char* name, const char* opts)
{
	GdamDeviceOss* dev_oss = GDAM_DEVICE_OSS(sink);
	int fd;
	GdamDeviceOptions options;

	if (!gdam_device_parse_options(&options, opts)) {
		g_warning("error parsing options `%s'", opts);
		return FALSE;
	}

	dev_oss->fragment_size = (1 << options.log_frag_size);

	fd = gdam_device_oss_open_fd(name, &options);
	if (fd < 0) {
		g_warning("error opening sound device %s", name);
		return FALSE;
	}
	dev_oss->device_fd = fd;
	dev_oss->downsample_func = options.downsample_func;
	dev_oss->sample_size = options.sample_size;
	dev_oss->has_written = FALSE;
	sink->device_name = g_strdup(name);
	gdam_debug("dev_oss: sample_size=%d", options.sample_size);

	if (sink->feed != NULL)
		gdam_device_oss_play(GDAM_DEVICE_OSS(sink));

	return TRUE;
}

/* Close the sound device. */
static void
gdam_device_oss_close(GdamSink* sink)
{
	GdamDeviceOss* dev_oss = GDAM_DEVICE_OSS(sink);
	gdam_device_oss_stop(dev_oss);
	if (dev_oss->device_fd >= 0) {
		close(dev_oss->device_fd);
		dev_oss->device_fd = -1;
	}
}

static gboolean
gdam_device_oss_set_source(GdamSink* sink, GdamSource* source)
{
	sink->feed = source;
	source->device = sink;
	gtk_object_ref(GTK_OBJECT(source));
	gdam_device_oss_play(GDAM_DEVICE_OSS(sink));
	return TRUE;
}

static void
gdam_device_oss_remove_source(GdamSink* sink)
{
	gdam_device_oss_stop(GDAM_DEVICE_OSS(sink));
	gtk_object_unref(GTK_OBJECT(sink->feed));
}

static void gdam_device_oss_destroy(GtkObject* to_kill)
{
	gdam_device_oss_stop(GDAM_DEVICE_OSS(to_kill));
	gdam_device_oss_close(GDAM_SINK(to_kill));
}

static int         gdam_device_oss_write     (GdamSink*         output,
	                                      const GdamSample* samples,
					      int               max_read)
{
	g_assert_not_reached();
	return -1;
}


/*
 * Adding and detaching from the main loop.
 */
static GSourceFuncs device_oss_source_funcs = {
	gdam_device_oss_prepare,
	gdam_device_oss_check,
	gdam_device_oss_dispatch,
	NULL
};
static void gdam_device_oss_source_destroy_notify(gpointer user_data);

static void gdam_device_oss_stop(GdamDeviceOss* device)
{
	gdam_debug("gdam_device_oss_stop...");
	if (device->poll_fd.fd >= 0) {
		g_main_remove_poll(&device->poll_fd);
		device->poll_fd.fd = -1;
	}
	if (device->source_id != 0) {
		guint source_id = device->source_id;
		device->source_id = 0;
		g_source_remove(source_id);
	}
}

static gboolean too_long_to_write_trigger(gpointer device_oss_ptr)
{
	GdamDeviceOss* device = GDAM_DEVICE_OSS(device_oss_ptr);
	const char* name = GDAM_SINK(device)->device_name;

	if (device->has_written
	 || GTK_OBJECT_DESTROYED(GTK_OBJECT(device))) {
		gtk_object_unref(GTK_OBJECT(device));
		return FALSE;
	}

	fprintf(stderr, "Hello...\n\n"
	                "You seem to be using a soundcard (%s), \n"
			"which doesn't support ``polling'', a feature\n"
			"GDAM uses extensively.  In order to use GDAM,\n"
			"you must specify the ``fork'' option to GDAM.\n"
			"\n"
			"Modify the server.config file, which is probably\n"
			"in %s, but you might also look in\n"
			"\t~/.gdam/server.config\n"
			"\t/usr/local/gdam/lib/server.config\n"
			"\t/usr/lib/server.config\n"
			"Are also places we look...\n\n"
			"The exact location should have been printed when we"
			"started.\n"
			"\nPlease modify the line that looks like\n"
			"\tDevice          \"oss\"           \"%s\"     \"\"\n"
			"to look like:\n"
			"\tDevice          \"oss\"           \"%s\"     "
						"\"fork\"\n"
			"(if there is already an option, put a colon in between,\n"
			"\t\"fork:numfragments=16\"\n"
			"or whatever...)\n"
			, name, GDAM_CONFIG_LOCATION, name, name);

	gtk_object_unref(GTK_OBJECT(device));
	return FALSE;
}

static void gdam_device_oss_play(GdamDeviceOss* device)
{
	g_return_if_fail(device->device_fd >= 0);
	g_return_if_fail(device->source_id == 0);
	g_return_if_fail(device->poll_fd.fd == -1);

	device->poll_fd.fd = device->device_fd;
	device->poll_fd.events = G_IO_OUT;
	g_main_add_poll(&device->poll_fd, GDAM_SOURCE_PRIORITY);

	device->source_id = g_source_add(
		GDAM_SOURCE_PRIORITY,
		TRUE,
		&device_oss_source_funcs,
		NULL,
		device,
		gdam_device_oss_source_destroy_notify
	);
	g_assert(device->source_id != 0);

	if (device->has_written == FALSE) {
		gtk_object_ref(GTK_OBJECT(device));
		g_timeout_add(3000,
		              too_long_to_write_trigger,
			      device);
	}
}

static void gdam_device_oss_source_destroy_notify(gpointer user_data)
{
	GdamDeviceOss* dev_oss = GDAM_DEVICE_OSS(user_data);
	dev_oss->source_id = 0;
	gdam_device_oss_stop(dev_oss);
}

static gboolean    gdam_device_oss_prepare   (gpointer          src_data,
                                              GTimeVal*         curtime,
                                              gint*             timeout,
		                              gpointer          user_data)
{
	*timeout = -1;
	return FALSE;
}

static gboolean    gdam_device_oss_check     (gpointer          src_data,
                                              GTimeVal*         curtime,
		                              gpointer          user_data)
{
	GdamDeviceOss* dev_oss = (GdamDeviceOss*)user_data;
	if (GDAM_SERVER_DEBUG_POLL)
		gdam_debug("gdam_device_oss_check: revents=%d", 
				dev_oss->poll_fd.revents);
	return dev_oss->poll_fd.revents == 0 ? FALSE : TRUE;
}

static gboolean    gdam_device_oss_dispatch  (gpointer          src_data,
                                              GTimeVal*         curtime,
		                              gpointer          user_data)
{
	if (!gdam_device_oss_write_one_fragment(GDAM_DEVICE_OSS(user_data)))
		return FALSE;
	return TRUE;
}

static gboolean
gdam_device_oss_write_one_fragment(GdamDeviceOss* device_oss)
{
	char* fragment_buffer;
	int num_samples = device_oss->fragment_size / sizeof(GdamSample);
	int num_read;
	GdamSink* sink = (GdamSink*)device_oss;
	GdamSource* feed = sink->feed;

	/* The fragment size is the ideal amount to write to
	 * an oss device atomically. (I think) */
	int fragment_size = device_oss->fragment_size;
	int out_fragment_size = fragment_size / sizeof(GdamSample)
	                      * device_oss->sample_size / 8;

	g_return_val_if_fail(feed != NULL, FALSE);

	fragment_buffer = (char*)alloca(fragment_size);


	/* Read, then miss. */
	if (!gdam_source_is_ready_to_read(feed))
		num_read = 0;
	else {
		num_read = gdam_source_read(feed,
		                            (GdamSample*)fragment_buffer,
					    num_samples);
		if (num_read <= 0) {
			gdam_debug("writing to device: subsource ended, "
			           "returned %d", num_read);

			/* Remove source sets feed to NULL. */
			gdam_sink_remove_source(sink);

			return FALSE;
		}
	}

	if (num_read < num_samples)
		gdam_source_miss_data(feed, num_samples - num_read);
	if (device_oss->downsample_func) {
		int num_samples = fragment_size / sizeof(GdamSample);
		(*device_oss->downsample_func)(fragment_buffer, num_samples);
	}

	device_oss->has_written = TRUE;

	for (;;) {
		int num_written = write(device_oss->device_fd,
		                        fragment_buffer,
				        out_fragment_size);
		if (num_written < 0) {
			if (errno == EINTR)
				continue;
			if (errno == EWOULDBLOCK) {
				g_warning("TODO: " /*TODO*/
				          "not able to handle non-ready, "
					  "non-blocking situation "
					  "(SHOULD QUEUE DATA)");
				g_warning("ditching a fragment of data");
				return TRUE;
			}
			g_warning("error writing to sound device: %s",
				  strerror(errno));

			/* Will set device_fd to -1. */
			gdam_sink_close(sink);

			return FALSE;
		}
		
		if (num_written < out_fragment_size) {
			static gboolean one_shot = FALSE;
			if (!one_shot) {
				g_warning("hmmm... your sound device didn't "
					  "accept a whole fragment...");
				g_warning("something may be wrong, but "
					  "we'll proceed");
				one_shot = TRUE;
			}
			g_assert(num_written != 0);
			fragment_size -= num_written;
			fragment_buffer += num_written;
			continue;
		}
		break;
	}

	return TRUE;
}
